# ==================
# Main file for ScaleBy
# by Alex Schreyer
# ==================


require 'sketchup.rb'


# ==================


module AS_Extensions

  module AS_ScaleBy
  
  
    # ==================
    
    
    # Helper function to convert a SketchUp color to greyscale (0..1)
    def self.grey( color )
      g = color.red.to_f / 255 * 0.2126 + 
          color.green.to_f / 255 * 0.7152 + 
          color.blue.to_f / 255 * 0.0722
      return g
    end  
    
    
    # ==================
    
    
    def self.scale_by_image
    # Scale objects based on image grayscale
    
        # Give warning where ImageRep isn't available
        if Sketchup.version.to_i < 18
            UI.messagebox "该工具仅适用于SketchUp 2018及更高版本。"
            return
        end    
    
        mod = Sketchup.active_model
        ent = mod.entities
        sel = mod.selection
        
        toolname = "通过图像变换对象"      
        
        # Get all objects from selection
        all_objects = []
        all_objects.push( *sel.grep( Sketchup::ComponentInstance ) )
        all_objects.push( *sel.grep( Sketchup::Group ) )
        
        if !all_objects.empty?
        
            # Get all the parameters from input dialog
            prompts = [ "要应用的变换(对象坐标)  " ,
                        "图像方向(本地坐标) " ,
                        "倍增(比例|角度|距离) " ]
            defaults = [ "统一比例" , "红轴-绿轴 (x-y)" , "2" ]
            lists = [ "统一比例|沿红轴缩放|沿绿轴缩放|沿蓝轴缩放|沿红轴旋转|沿绿轴旋转|沿蓝轴旋转|沿红轴移动|沿绿轴移动|沿蓝轴移动" , 
                      "红轴-绿轴 (x-y)|红轴-蓝轴 (x-z)|绿轴-蓝轴 (y-z)" , "" ]
            defaults = Sketchup.read_default( @extname , __method__.to_s , defaults )
            
            res = UI.inputbox( prompts , defaults , lists , toolname )
            return if !res
            
            Sketchup.write_default( @extname , __method__.to_s , res.map { |s| s.gsub( '"' , '' ) } )  # Fix for inch pref saving error
            
            mod.start_operation toolname
            
            begin
            
                if res[0].include? "移动"
                    s_fac = res[2].to_l 
                else
                    s_fac = res[2].to_f
                end
                
                # Ask for an image file
                f = UI.openpanel '选择一个图像文件', '', '图像文件|*.jpg;*.png;||'
                return if f == nil

                # Now load the image
                ir = Sketchup::ImageRep.new
                ir.load_file( f )                

                # Create temporary group from selection to get dimensions
                gr = ent.add_group( all_objects )
                gr_x_dim = ( gr.definition.bounds.max - gr.definition.bounds.min ).x
                gr_y_dim = ( gr.definition.bounds.max - gr.definition.bounds.min ).y
                gr_z_dim = ( gr.definition.bounds.max - gr.definition.bounds.min ).z 

                # Work with entities in group
                gr.entities.each_with_index { |e,i|
                
                    # Get center of object
                    if e.is_a? Sketchup::ComponentInstance
                        cen = e.transformation.origin
                    else
                        cen = e.bounds.center
                    end
                    
                    # Scale based on image orientation
                    if res[1] == "红轴-蓝轴 (x-z)"  # Front orientation
                        
                        x = cen.x.to_f / gr_x_dim
                        z = cen.z.to_f / gr_z_dim

                        scale = self.grey( ir.color_at_uv( x , z , true ) )     
                    
                    elsif res[1] == "绿轴-蓝轴 (y-z)"  # Side orientation
                        
                        y = cen.y.to_f / gr_y_dim
                        z = cen.z.to_f / gr_z_dim

                        scale = self.grey( ir.color_at_uv( y , z , true ) )      
                    
                    else  # Flat on the ground

                        x = cen.x.to_f / gr_x_dim
                        y = cen.y.to_f / gr_y_dim

                        scale = self.grey( ir.color_at_uv( x , y , true ) )     
                    
                    end       
                    
                    # Get transformation                             
                    if res[0] == "沿红轴缩放"

                        t = Geom::Transformation.scaling scale * s_fac , 1 , 1

                    elsif res[0] == "沿绿轴缩放"

                        t = Geom::Transformation.scaling 1 , scale * s_fac , 1

                    elsif res[0] == "沿蓝轴缩放"

                        t = Geom::Transformation.scaling 1 , 1 , scale * s_fac                    

                    elsif res[0] == "沿红轴旋转"

                        t = Geom::Transformation.rotation [ 0 , 0 , 0 ] , [ 1 , 0 , 0 ] , ( scale * s_fac ).degrees

                    elsif res[0] == "沿绿轴旋转"

                        t = Geom::Transformation.rotation [ 0 , 0 , 0 ] , [ 0 , 1 , 0 ] , ( scale * s_fac ).degrees

                    elsif res[0] == "沿蓝轴旋转"

                        t = Geom::Transformation.rotation [ 0 , 0 , 0 ] , [ 0 , 0 , 1 ] , ( scale * s_fac ).degrees

                    elsif res[0] == "沿红轴移动"

                        t = Geom::Transformation.translation [ scale * s_fac , 0 , 0 ]

                    elsif res[0] == "沿绿轴移动"

                        t = Geom::Transformation.translation [ 0 , scale * s_fac , 0 ]

                    elsif res[0] == "沿蓝轴移动"

                        t = Geom::Transformation.translation [ 0 , 0 , scale * s_fac ]                        

                    else

                        # Uniform scaling as default
                        t = Geom::Transformation.scaling scale * s_fac

                    end

                    # Apply the transformation
                    e.transformation *= t
                    
                    # Life is always better with some feedback while SketchUp works
                    Sketchup.status_text = toolname + " | 完成对象 #{(i+1).to_s}"

                }

                # Explode group once we are done
                gr.explode            
            
            rescue Exception => e    
            
                UI.messagebox("无法完成！错误: #{e}")
                
            end
            
            mod.commit_operation
            
        else  # Can't start tool
        
            UI.messagebox "在使用此工具前请先选择要进行操作的多个对象(组或组件实例)。"
        
        end               
    
    end  # scale_by_image
    
    
    # ==================
    
    
    def self.scale_by_attractor
    # Scale objects based on proximity to attractors
    
        mod = Sketchup.active_model
        ent = mod.entities
        sel = mod.selection
        
        toolname = "通过热点干扰变换对象"      
        
        # Get all objects from selection
        all_objects = []
        all_objects.push( *sel.grep( Sketchup::ComponentInstance ) )
        all_objects.push( *sel.grep( Sketchup::Group ) )
        
        # Get attractors from selection
        att = all_objects.grep( Sketchup::ComponentInstance ) { |c| c if c.definition.name == "A" }
        att.compact!
        
        all_objects = all_objects - att
        
        if !( all_objects.empty? or att.empty? )
        
            # Get all the parameters from input dialog
            prompts = [ "要应用的变换(对象坐标) " ,
                        "吸引力比例 (距离) " ,
                        "吸引力衰减 " ,
                        "吸引力类型 " ,
                        "倍增(比例|角度|距离) " ]
            defaults = [ "统一比例" , "10'" , "线性" , "减少" , "2" ]
            lists = [ "统一比例|沿红轴缩放|沿绿轴缩放|沿蓝轴缩放|沿红轴旋转|沿绿轴旋转|沿蓝轴旋转|沿红轴移动|沿绿轴移动|沿蓝轴移动" , 
                      "" , "平方根|线性|平方" , "减少|增加" , "" ]
            defaults = Sketchup.read_default( @extname , __method__.to_s , defaults )
            
            res = UI.inputbox( prompts , defaults , lists , toolname )
            return if !res
            
            Sketchup.write_default( @extname , __method__.to_s , res.map { |s| s.gsub( '"' , '' ) } )  # Fix for inch pref saving error
            
            mod.start_operation toolname
            
            begin
            
                # Get parameters
                att_val = res[1].to_l 
                if res[2] == "平方根"
                    power = 0.5
                elsif res[2] == "平方"
                    power = 2
                else
                    power = 1
                end  
                if res[0].include? "移动"
                    s_fac = res[4].to_l 
                else
                    s_fac = res[4].to_f
                end

                # Create temporary group from selection to get dimensions
                gr = ent.add_group( all_objects + att )
                gr_x_dim = ( gr.definition.bounds.max - gr.definition.bounds.min ).x
                gr_y_dim = ( gr.definition.bounds.max - gr.definition.bounds.min ).y
                gr_z_dim = ( gr.definition.bounds.max - gr.definition.bounds.min ).z 

                # Work with entities in group
                gr.entities.each_with_index { |e,i|
                
                    if !att.include?(e)
                
                        # Get center of object
                        if e.is_a? Sketchup::ComponentInstance
                            cen = e.transformation.origin
                        else
                            cen = e.bounds.center
                        end
                        
                        # Find the smallest attractor distance, start with large number
                        dist = 1000000
                        
                        att.each { |a|
                        
                            a_dist = a.transformation.origin.distance cen
                            dist = a_dist if a_dist < dist
                        
                        }
                        
                        scale = ( dist**power / att_val**power ).to_f
                        
                        # Adjust up/down scaling direction                        
                        scale = 1 / scale if res[3] == "增加"

                        # Get transformation                             
                        if res[0] == "沿红轴缩放"

                            t = Geom::Transformation.scaling scale * s_fac , 1 , 1

                        elsif res[0] == "沿绿轴缩放"

                            t = Geom::Transformation.scaling 1 , scale * s_fac , 1

                        elsif res[0] == "沿蓝轴缩放"

                            t = Geom::Transformation.scaling 1 , 1 , scale * s_fac                    

                        elsif res[0] == "沿红轴旋转"

                            t = Geom::Transformation.rotation [ 0 , 0 , 0 ] , [ 1 , 0 , 0 ] , ( scale * s_fac ).degrees

                        elsif res[0] == "沿绿轴旋转"

                            t = Geom::Transformation.rotation [ 0 , 0 , 0 ] , [ 0 , 1 , 0 ] , ( scale * s_fac ).degrees

                        elsif res[0] == "沿蓝轴旋转"

                            t = Geom::Transformation.rotation [ 0 , 0 , 0 ] , [ 0 , 0 , 1 ] , ( scale * s_fac ).degrees

                        elsif res[0] == "沿红轴移动"

                            t = Geom::Transformation.translation [ scale * s_fac , 0 , 0 ]

                        elsif res[0] == "沿绿轴移动"

                            t = Geom::Transformation.translation [ 0 , scale * s_fac , 0 ]

                        elsif res[0] == "沿蓝轴移动"

                            t = Geom::Transformation.translation [ 0 , 0 , scale * s_fac ]                        

                        else

                            # Uniform scaling as default
                            t = Geom::Transformation.scaling scale * s_fac

                        end

                        # Apply the transformation
                        e.transformation *= t

                        # Life is always better with some feedback while SketchUp works
                        Sketchup.status_text = toolname + " | 完成对象 #{(i+1).to_s}"
                        
                    end

                }

                # Explode group once we are done
                gr.explode            
            
            rescue Exception => e    
            
                UI.messagebox("无法完成！错误: #{e}")
                
            end
            
            mod.commit_operation
            
        else  # Can't start tool
        
            UI.messagebox "在使用此工具前请先选择要进行操作的多个对象（组或组件实例）以及一个或多个名为“ A”的组件。"
        
        end               
    
    end  # scale_by_attractor    
    
    
    # ==================
    
    
    def self.pushpull_by_image
    # Push/pull faces based on image grayscale
    
        # Give warning where ImageRep isn't available
        if Sketchup.version.to_i < 18
            UI.messagebox "该工具仅适用于SketchUp 2018及更高版本。"
            return
        end    
    
        mod = Sketchup.active_model
        ent = mod.entities
        sel = mod.selection
        
        toolname = "按图像推/拉面"     
        
        # Get all faces from selection
        all_faces = sel.grep( Sketchup::Face )
        
        if !all_faces.empty?
        
            # Get all the parameters from input dialog
            prompts = [ "最小推拉距离 " , 
                        "最大推拉距离 " , 
                        "创建新面 " ,
                        "图像方向(本地坐标) " ]
            defaults = [ "0" , "1'" , "是" , "红轴-绿轴 (x-y)" ]
            lists = [ "" , "" , "是|否" , "红轴-绿轴 (x-y)|红轴-蓝轴 (x-z)|绿轴-蓝轴 (y-z)" ]
            defaults = Sketchup.read_default( @extname , __method__.to_s , defaults )
            
            res = UI.inputbox( prompts , defaults , lists , toolname )
            return if !res
            
            Sketchup.write_default( @extname , __method__.to_s , res.map { |s| s.gsub( '"' , '' ) } )  # Fix for inch pref saving error
            
            mod.start_operation toolname
            
            begin
            
                # Get extrusion distances and convert to length
                min_pp = res[0].to_l
                max_pp = res[1].to_l

                # Ask for an image file
                f = UI.openpanel '选择一个图像文件', '', '图像文件|*.jpg;*.png;||'
                return if f == nil

                # Now load the image
                ir = Sketchup::ImageRep.new
                ir.load_file( f )

                # Create temporary group from selection to get dimensions
                gr = ent.add_group( all_faces )
                gr_x_dim = ( gr.definition.bounds.max - gr.definition.bounds.min ).x
                gr_y_dim = ( gr.definition.bounds.max - gr.definition.bounds.min ).y
                gr_z_dim = ( gr.definition.bounds.max - gr.definition.bounds.min ).z

                # Work with all faces
                all_faces.each_with_index { |e,i|
                
                    # Get center of face
                    cen = e.bounds.center
                    
                    # Scale based on orientation
                    if res[3] == "红轴-蓝轴 (x-z)"  # Front orientation
                    
                        x = cen.x.to_f / gr_x_dim
                        z = cen.z.to_f / gr_z_dim

                        scale = self.grey( ir.color_at_uv( x , z , true ) )                     
                    
                    elsif res[3] == "绿轴-蓝轴 (y-z)"  # Side orientation
                    
                        y = cen.y.to_f / gr_y_dim
                        z = cen.z.to_f / gr_z_dim

                        scale = self.grey( ir.color_at_uv( y , z , true ) )     
                    
                    else  # Flat on the ground
                    
                        x = cen.x.to_f / gr_x_dim
                        y = cen.y.to_f / gr_y_dim

                        scale = self.grey( ir.color_at_uv( x , y , true ) )      
                    
                    end
                    
                    # Extrude face
                    e.pushpull( min_pp + scale * ( max_pp - min_pp ) , res[2] == "是" ? true : false )      
                    
                    # Life is always better with some feedback while SketchUp works
                    Sketchup.status_text = toolname + " | 完成面 #{(i+1).to_s}"

                }     
                
                # Explode group once we are done
                gr.explode                 
            
            rescue Exception => e    
            
                UI.messagebox("无法完成！错误: #{e}")
                
            end 
            
            mod.commit_operation
            
        else  # Can't start tool
        
            UI.messagebox "在使用此工具前请先选择要进行操作的多个面。 然后在弹出的窗口确定参数后将会要求您选择图片。"
        
        end            
    
    end  # pushpull_by_image    
    
    
    # ==================
    
    
    def self.vertices_by_image
    # Move vertices based on image grayscale
    
        # Give warning where ImageRep isn't available
        if Sketchup.version.to_i < 18
            UI.messagebox "该工具仅适用于SketchUp 2018及更高版本。"
            return
        end    
    
        mod = Sketchup.active_model
        ent = mod.entities
        sel = mod.selection
        
        toolname = "按图像移动顶点"      
        
        # Get all selected edges (and faces if we have some)
        all_edges = mod.selection.grep( Sketchup::Edge )
        all_faces = mod.selection.grep( Sketchup::Face )
        
        if !all_edges.empty?
        
            # Get all the parameters from input dialog
            prompts = [ "沿红轴最大移动距离 (x 距离) " , 
                        "沿绿轴最大移动距离 (y 距离) " , 
                        "沿蓝轴最大移动距离 (z 距离) " ,
                        "图像方向 (全为局部坐标) " ]
            defaults = [ "0" , "0" , "1'" , "红轴-绿轴 (x-y)" ]
            lists = [ "" , "" , "" , "红轴-绿轴 (x-y)|红轴-蓝轴 (x-z)|绿轴-蓝轴 (y-z)" ]
            defaults = Sketchup.read_default( @extname , __method__.to_s , defaults )
            
            res = UI.inputbox( prompts , defaults , lists , toolname )
            return if !res
            
            Sketchup.write_default( @extname , __method__.to_s , res.map { |s| s.gsub( '"' , '' ) } )  # Fix for inch pref saving error
            
            mod.start_operation toolname
            
            begin
            
                # Get max distances and convert to length
                max_x = res[0].to_l
                max_y = res[1].to_l
                max_z = res[2].to_l

                # Ask for an image file
                f = UI.openpanel '选择一个图像文件', '', '图像文件|*.jpg;*.png;||'
                return if f == nil

                # Now load the image
                ir = Sketchup::ImageRep.new
                ir.load_file( f )
             
                # Create temporary group from selection to get dimensions
                gr = ent.add_group( all_edges + all_faces ) 
                gr_x_dim = ( gr.definition.bounds.max - gr.definition.bounds.min ).x
                gr_y_dim = ( gr.definition.bounds.max - gr.definition.bounds.min ).y
                gr_z_dim = ( gr.definition.bounds.max - gr.definition.bounds.min ).z    
           
                # Get all the unique vertices                
                gr_edges = gr.entities.grep( Sketchup::Edge )
                vertices = []
                gr_edges.each { |e| vertices << e.vertices }
                vertices.flatten!
                vertices.uniq!

                vertices.each_with_index { |v,i| 
                
                    # Scale based on position and image grey value
                    cen = v.position
                    
                    # Scale based on orientation
                    if res[3] == "红轴-蓝轴 (x-z)"  # Front orientation
                    
                        x = cen.x.to_f / gr_x_dim
                        z = cen.z.to_f / gr_z_dim

                        scale = self.grey( ir.color_at_uv( x , z , true ) )                     
                    
                    elsif res[3] == "绿轴-蓝轴 (y-z)"  # Side orientation
                    
                        y = cen.y.to_f / gr_y_dim
                        z = cen.z.to_f / gr_z_dim

                        scale = self.grey( ir.color_at_uv( y , z , true ) )    
                    
                    else  # Flat on the ground
                    
                        x = cen.x.to_f / gr_x_dim
                        y = cen.y.to_f / gr_y_dim

                        scale = self.grey( ir.color_at_uv( x , y , true ) )  
                    
                    end

                    t = Geom::Transformation.new [  scale * max_x , scale * max_y , scale * max_z ]
                    gr.entities.transform_entities( t , v )
                    
                    # Life is always better with some feedback while SketchUp works
                    Sketchup.status_text = toolname + " | 完成顶点 #{(i+1).to_s}"
                    
                }
                
                # Explode group once we are done
                gr.explode  
            
            rescue Exception => e    
            
                UI.messagebox("无法完成！错误: #{e}")
                
            end
            
            mod.commit_operation
            
        else  # Can't start tool
        
            UI.messagebox "在使用此工具前请先选择要进行操作的多条边线（如果连接、则选择面）（它们的顶点将被移动）。 然后在弹出的窗口确定参数后将会要求您选择图片。"
        
        end            
    
    end  # vertices_by_image    
    
    
    # ==================
    
    
    def self.scale_by_math_power
    # Scale objects based on a math formula
    
        mod = Sketchup.active_model
        ent = mod.entities
        sel = mod.selection
        
        toolname = "通过幂方程式变换对象"      
        
        # Get all objects from selection
        all_objects = []
        all_objects.push( *sel.grep( Sketchup::ComponentInstance ) )
        all_objects.push( *sel.grep( Sketchup::Group ) )
        
        if !all_objects.empty?
        
            # Get all the parameters from input dialog
            prompts = [ "要应用的变换(对象坐标)  " ,
                        "红轴倍增 (x) (A in f = A(Bx)^C+D) " , 
                        "红轴功率因数 (x) (B in f = A(Bx)^C+D) " ,
                        "红轴的幂 (x) (C in f = A(Bx)^C+D) " ,
                        "绿轴倍增 (y) (A in f = A(By)^C+D) " , 
                        "绿轴功率因数 (y) (B in f = A(By)^C+D) " ,
                        "绿轴的幂 (y) (C in f = A(By)^C+D) " ,
                        "蓝轴倍增 (z) (A in f = A(Bz)^C+D) " , 
                        "蓝轴功率因数 (z) (B in f = A(Bz)^C+D) " ,
                        "蓝轴的幂 (z) (C in f = A(Bz)^C+D) " ,
                        "偏移量 (D) (全为局部坐标) " ]
            defaults = [ "统一比例" , "1" , "2" , "2" , "1" , "2" , "2" , "0" , "0" , "0" , "0" ]
            lists = [ "统一比例|沿红轴缩放|沿绿轴缩放|沿蓝轴缩放|沿红轴旋转|沿绿轴旋转|沿蓝轴旋转|沿红轴移动|沿绿轴移动|沿蓝轴移动" , "" , "" , "" , "" , "" , "" , "" , "" , "" , "" ]
            defaults = Sketchup.read_default( @extname , __method__.to_s , defaults )
            
            res = UI.inputbox( prompts , defaults , lists , toolname )
            return if !res
            
            Sketchup.write_default( @extname , __method__.to_s , res.map { |s| s.gsub( '"' , '' ) } )  # Fix for inch pref saving error
            
            mod.start_operation toolname
            
            begin
            
                x_fac = res[1].to_f
                y_fac = res[4].to_f
                z_fac = res[7].to_f
                x_pfa = res[2].to_f
                y_pfa = res[5].to_f
                z_pfa = res[8].to_f                
                x_pow = res[3].to_f
                y_pow = res[6].to_f
                z_pow = res[9].to_f
                off = res[10].to_f

                # Create temporary group from selection to get dimensions
                gr = ent.add_group( all_objects )
                gr_x_dim = ( gr.definition.bounds.max - gr.definition.bounds.min ).x
                gr_y_dim = ( gr.definition.bounds.max - gr.definition.bounds.min ).y
                gr_z_dim = ( gr.definition.bounds.max - gr.definition.bounds.min ).z 

                # Work with entities in group
                gr.entities.each_with_index { |e,i|
                
                    # Get base of object
                    if e.is_a? Sketchup::ComponentInstance
                        cen = e.transformation.origin
                    else
                        cen = e.bounds.center
                    end
                    
                    # Get combined scale
                    scale = 0.0
                    scale += x_fac * ( x_pfa * ( cen.x / gr_x_dim - 0.5)) ** x_pow  if gr_x_dim.abs > 0
                    scale += y_fac * ( y_pfa * ( cen.y / gr_y_dim - 0.5)) ** y_pow  if gr_y_dim.abs > 0
                    scale += z_fac * ( z_pfa * ( cen.z / gr_z_dim - 0.5)) ** z_pow  if gr_z_dim.abs > 0
                    scale += off
                    
                    # Get transformation                             
                    if res[0] == "沿红轴缩放"

                        t = Geom::Transformation.scaling scale , 1 , 1

                    elsif res[0] == "沿绿轴缩放"

                        t = Geom::Transformation.scaling 1 , scale , 1

                    elsif res[0] == "沿蓝轴缩放"

                        t = Geom::Transformation.scaling 1 , 1 , scale                    

                    elsif res[0] == "沿红轴旋转"

                        t = Geom::Transformation.rotation [ 0 , 0 , 0 ] , [ 1 , 0 , 0 ] , ( scale ).degrees

                    elsif res[0] == "沿绿轴旋转"

                        t = Geom::Transformation.rotation [ 0 , 0 , 0 ] , [ 0 , 1 , 0 ] , ( scale ).degrees

                    elsif res[0] == "沿蓝轴旋转"

                        t = Geom::Transformation.rotation [ 0 , 0 , 0 ] , [ 0 , 0 , 1 ] , ( scale ).degrees

                    elsif res[0] == "沿红轴移动"

                        t = Geom::Transformation.translation [ scale , 0 , 0 ]

                    elsif res[0] == "沿绿轴移动"

                        t = Geom::Transformation.translation [ 0 , scale , 0 ]

                    elsif res[0] == "沿蓝轴移动"

                        t = Geom::Transformation.translation [ 0 , 0 , scale ]                        

                    else

                        # Uniform scaling as default
                        t = Geom::Transformation.scaling scale

                    end

                    # Apply the transformation
                    e.transformation *= t                    
                    
                    # Life is always better with some feedback while SketchUp works
                    Sketchup.status_text = toolname + " | 已完成对象 #{(i+1).to_s}"

                }

                # Explode group once we are done
                gr.explode            
            
            rescue Exception => e    
            
                UI.messagebox("无法完成！错误: #{e}")
                
            end
            
            mod.commit_operation
            
        else  # Can't start tool
        
            UI.messagebox "在使用此工具前请先选择要进行操作的多个对象(组或组件实例)。"
        
        end            
    
    end  # scale_by_math_power  
    
    
    # ==================
    
    
    def self.scale_by_math_sine
    # Scale objects based on a math formula
    
        mod = Sketchup.active_model
        ent = mod.entities
        sel = mod.selection
        
        toolname = "通过正弦/余弦方程变换对象"      
        
        # Get all objects from selection
        all_objects = []
        all_objects.push( *sel.grep( Sketchup::ComponentInstance ) )
        all_objects.push( *sel.grep( Sketchup::Group ) )
        
        if !all_objects.empty?
        
            # Get all the parameters from input dialog
            prompts = [ "要应用的变换（对象坐标） " ,
                        "使用 " ,
                        "红轴振幅 (x) (A in f = A*sin(Bx)+D) " , 
                        "红轴周期 (x) (B in f = A*sin(Bx)+D) " ,
                        "绿轴振幅 (y) (A in f = A*sin(By)+D) " , 
                        "绿轴周期 (y) (B in f = A*sin(By)+D) " ,
                        "蓝轴振幅 (z) (A in f = A*sin(Bz)+D) " , 
                        "蓝轴周期 (z) (B in f = A*sin(Bz)+D) " ,
                        "偏移量 (D) (全为局部坐标) " ]
            defaults = [ "统一比例" , "正弦" , "1" , "1" , "1" , "1" , "0" , "0" , "0" ]
            lists = [ "统一比例|沿红轴缩放|沿绿轴缩放|沿蓝轴缩放|沿红轴旋转|沿绿轴旋转|沿蓝轴旋转|沿红轴移动|沿绿轴移动|沿蓝轴移动" , "正弦|余弦" , "" , "" , "" , "" , "" , "" , "" , "" , "" , "" ]
            defaults = Sketchup.read_default( @extname , __method__.to_s , defaults )
            
            res = UI.inputbox( prompts , defaults , lists , toolname )
            return if !res
            
            Sketchup.write_default( @extname , __method__.to_s , res.map { |s| s.gsub( '"' , '' ) } )  # Fix for inch pref saving error
            
            mod.start_operation toolname
            
            begin
            
                x_fac = res[2].to_f
                y_fac = res[4].to_f
                z_fac = res[6].to_f
                x_pfa = res[3].to_f
                y_pfa = res[5].to_f
                z_pfa = res[7].to_f                
                off = res[8].to_f

                # Create temporary group from selection to get dimensions
                gr = ent.add_group( all_objects )
                gr_x_dim = ( gr.definition.bounds.max - gr.definition.bounds.min ).x
                gr_y_dim = ( gr.definition.bounds.max - gr.definition.bounds.min ).y
                gr_z_dim = ( gr.definition.bounds.max - gr.definition.bounds.min ).z 

                # Work with entities in group
                gr.entities.each_with_index { |e,i|
                
                    # Get base of object
                    if e.is_a? Sketchup::ComponentInstance
                        cen = e.transformation.origin
                    else
                        cen = e.bounds.center
                    end
                    
                    # Get combined scale
                    if res[1] == "余弦"
                    
                        scale = 0.0
                        scale += x_fac * Math::cos( x_pfa * Math::PI * ( cen.x / gr_x_dim - 0.5)) if gr_x_dim.abs > 0
                        scale += y_fac * Math::cos( y_pfa * Math::PI * ( cen.y / gr_y_dim - 0.5)) if gr_y_dim.abs > 0
                        scale += z_fac * Math::cos( z_pfa * Math::PI * ( cen.z / gr_z_dim - 0.5)) if gr_z_dim.abs > 0
                        scale += off                        
                    
                    else
                    
                        scale = 0.0
                        scale += x_fac * Math::sin( x_pfa * Math::PI * ( cen.x / gr_x_dim - 0.5)) if gr_x_dim.abs > 0
                        scale += y_fac * Math::sin( y_pfa * Math::PI * ( cen.y / gr_y_dim - 0.5)) if gr_y_dim.abs > 0
                        scale += z_fac * Math::sin( z_pfa * Math::PI * ( cen.z / gr_z_dim - 0.5)) if gr_z_dim.abs > 0
                        scale += off    
                    
                    end
                    
                    # Get transformation                             
                    if res[0] == "沿红轴缩放"

                        t = Geom::Transformation.scaling scale , 1 , 1

                    elsif res[0] == "沿绿轴缩放"

                        t = Geom::Transformation.scaling 1 , scale , 1

                    elsif res[0] == "沿蓝轴缩放"

                        t = Geom::Transformation.scaling 1 , 1 , scale                    

                    elsif res[0] == "沿红轴旋转"

                        t = Geom::Transformation.rotation [ 0 , 0 , 0 ] , [ 1 , 0 , 0 ] , ( scale ).degrees

                    elsif res[0] == "沿绿轴旋转"

                        t = Geom::Transformation.rotation [ 0 , 0 , 0 ] , [ 0 , 1 , 0 ] , ( scale ).degrees

                    elsif res[0] == "沿蓝轴旋转"

                        t = Geom::Transformation.rotation [ 0 , 0 , 0 ] , [ 0 , 0 , 1 ] , ( scale ).degrees

                    elsif res[0] == "沿红轴移动"

                        t = Geom::Transformation.translation [ scale , 0 , 0 ]

                    elsif res[0] == "沿绿轴移动"

                        t = Geom::Transformation.translation [ 0 , scale , 0 ]

                    elsif res[0] == "沿蓝轴移动"

                        t = Geom::Transformation.translation [ 0 , 0 , scale ]                        

                    else

                        # Uniform scaling as default
                        t = Geom::Transformation.scaling scale

                    end

                    # Apply the transformation
                    e.transformation *= t
                    
                    # Life is always better with some feedback while SketchUp works
                    Sketchup.status_text = toolname + " | 已完成对象 #{(i+1).to_s}"

                }

                # Explode group once we are done
                gr.explode            
            
            rescue Exception => e    
            
                UI.messagebox("无法完成！错误: #{e}")
                
            end
            
            mod.commit_operation
            
        else  # Can't start tool
        
            UI.messagebox "在使用此工具前请先选择要进行操作的多个对象(组或组件实例)。"
        
        end            
    
    end  # scale_by_math_sine    


    # ==================
    
    
    def self.show_url( title , url )
    # Show website either as a WebDialog or HtmlDialog
    
      if Sketchup.version.to_f < 17 then   # Use old dialog
        @dlg = UI::WebDialog.new( title , true ,
          title.gsub(/\s+/, "_") , 1000 , 600 , 100 , 100 , true);
        @dlg.navigation_buttons_enabled = false
        @dlg.set_url( url )
        @dlg.show      
      else   #Use new dialog
        @dlg = UI::HtmlDialog.new( { :dialog_title => title, :width => 1000, :height => 600,
          :style => UI::HtmlDialog::STYLE_DIALOG, :preferences_key => title.gsub(/\s+/, "_") } )
        @dlg.set_url( url )
        @dlg.show
        @dlg.center
      end  
    
    end    


    def self.show_help
    # Browse news using webdialog
    
      show_url( @exttitle , "https://alexschreyer.net/projects/scale-by-image-equation-tools-extension-for-sketchup/" ) 

    end # show_help


    # ==================


    if !file_loaded?(__FILE__)

        tools = []
        tools << [ "通过图像变换对象" , "scale_by_image" , "选择多个群组或组件，运行命令后选择图片，等待计算完成后将按参数根据图像的色阶、明暗自动缩放，整体形成图像的内容。" ]
        tools << [ "通过热点干扰变换对象" , "scale_by_attractor" , "选择多个群组或组件，和一个或几个命名为“A”的组件引力源，运行命令后设置幂公式参数，对象组将根据引力源的位置自动缩放形成参数所指的渐变效果。" ]      
        tools << [ "通过幂方程式变换对象" , "scale_by_math_power" , "选择多个群组或组件，运行命令后设置幂公式参数，对象组将自动缩放形成参数所指的渐变效果。" ]
        tools << [ "通过正弦/余弦方程变换对象" , "scale_by_math_sine" , "选择多个群组或组件，运行命令后设置正弦/余弦公式参数，对象组将自动缩放形成参数所指的渐变效果。" ]
        tools << [ "" , "" , "" ]
        tools << [ "按图像推/拉面" , "pushpull_by_image" , "选择多个面，运行命令后选择图片，等待计算完成后将按参数根据图像的色阶、明暗自动推拉面，整体形成图像的内容。" ]
        tools << [ "按图像移动点" , "vertices_by_image" , "选择多条边和面（已连接形成了顶点），运行命令后选择图片，等待计算完成后将按参数根据图像的色阶、明暗移动顶点，整体形成图像的内容。" ]

        # Add to the SketchUp Extensions menu and create a toolbar
        menu = UI.menu( "Tools" ).add_submenu( @exttitle )
        toolbar = UI::Toolbar.new @exttitle 
        
        # Get icon file extension
        sm = lg = ""    
        RUBY_PLATFORM =~ /darwin/ ? ext = "pdf" : ext = "svg"
        if Sketchup.version.to_i < 16  
            ext = "png"
            sm = "_sm"
            lg = "_lg"
        end     
        
        # Add them all to menu and toolbar
        tools.each { |t|
        
            if ( t[0] != "" )

                cmd = UI::Command.new( t[0] ) { self.send( t[1] ) }
                cmd.small_icon = File.join( @extdir , @extname , "icons" , t[1] + "#{sm}.#{ext}")
                cmd.large_icon = File.join( @extdir , @extname , "icons" , t[1] + "#{lg}.#{ext}")
                cmd.tooltip = t[0]
                cmd.status_bar_text = t[2]
                menu.add_item cmd
                toolbar.add_item cmd    
                
            else
            
                menu.add_separator
                toolbar.add_separator
            
            end

        }        

        # And a link to get help only to the menu
        menu.add_separator
        menu.add_item( "帮助" ) { self.show_help }
        
        
        # Don't forget to show the toolbar
        toolbar.show

        # Let Ruby know we have loaded this file
        file_loaded(__FILE__)

    end # if


    # ==================


  end # module AS_ScaleBy

end # module AS_Extensions


# ==================
